import math
import random
from game import Game
from typing import Callable


#
# Data types which might be useful for your implementation
# For more details follow the seminar slides
#
class Node(object):
    def __init__(self):
        self.visit_count = 0
        self.to_play = 0
        self.value_sum = 0
        self.children = {}
        self.unexplored_actions = None
        self.is_terminal = False

    def expanded(self) -> bool:
        return len(self.children) > 0 or self.is_terminal

    def fully_expanded(self) -> bool:
        return self.expanded() and \
               self.unexplored_actions is not None and \
               len(self.unexplored_actions) == 0 and \
               not self.is_terminal

    def value(self) -> float:
        if self.visit_count == 0:
            return 0
        return self.value_sum / self.visit_count


class DeterministicAgent:
    def __init__(self, evaluate_fn: Callable[[Game, int], float], maximizing_player: int, depth: int):
        self.maximizing_player = maximizing_player
        self.evaluate_fn = evaluate_fn
        self.estimate_depth = depth

    def compute_estimate(self, game: Game) -> float:
        """
        Computes the game estimate
        """
        assert game.to_play == self.maximizing_player
        return self.alphabeta(game, -float("inf"), float("inf"), self.estimate_depth, 1)

    def act(self, game: Game) -> int:
        """
        Selects the best action according to its estimate
        """

        def _step_new(g, action):
            g = g.clone()
            g.apply(action)
            return g

        assert game.to_play == self.maximizing_player
        _, action = max(((self.compute_estimate(_step_new(game, action)), action) for action in game.legal_actions()))
        return action

    def alphabeta(self, game: Game, alpha: float, beta: float, depth: int, color: int) -> float:
        #
        # TODO: implement
        # Your goal is to minimize the number of expanded nodes (calls to the legal_actions function).
        #

        # Implementation hints:
        # - In some games you can get to the same states by following different sequence of actions. Caching states
        # can be useful.
        # - Function estimate_fn has the following form: (game, player) -> float, where :game is the game instance
        #  and :player is the player whose score you want to compute. Since the game is a zero-sum game, the other
        #  player has score of minus that value.
        # - It is recommended to call estimate_fn always with self.maximizing_player
        # - Players switch sides after each move!
        # - We recommend you to use color 1 for maximizing player and -1 for minimizing player
        return 0


class MonteCarloAgent:
    def __init__(self, num_simulations: int, uct_constant: float = math.sqrt(2)):
        self.num_simulations = num_simulations
        self.uct_constant = uct_constant

    def compute_estimate(self, game: Game) -> float:
        """
        Computes the game estimate
        """
        root = self.mcts(game)
        return root.value_sum / root.visit_count

    def act(self, game: Game) -> int:
        root = self.mcts(game)
        return self._select_action(root)

    def mcts(self, game: Game) -> Node:
        """
        Runs MCTS search on current game state
        """
        root = Node()
        for _ in range(self.num_simulations):
            #
            # TODO: implement
            #
            pass
        return root

    def _select_action(self, node):
        _, action = max((c.visit_count, a) for a, c in node.items())
        return action

    def _expand(self, search_path, node, game):
        """
        If the node is terminal, does nothing and returns the node.
        Otherwise, expands the node, selects a child, adds the child on the search path and returns it.
        """
        #
        # TODO: implement
        #
        pass

    def _backpropagate(self, search_path, value):
        """
        At the end of a simulation, we propagate the evaluation all the way up the
        tree to the root. We also increase the visit_count.
        :param float value: The game score relative to the last state on the search path. It could be 1 for winning player and -1 for losing player.
        """
        #
        # TODO: implement
        #
        # Note: when we backpropagate to a state with different to_play, we have to negate the result
        pass

    def _simulate(self, node, game):
        """
        We simulate one rollout starting at node :node with current game state.
        :return: terminal_value relative to the node.to_play player.
        """
        #
        # TODO: implement
        #
        pass

    def _uct_select(self, node):
        _, action, child = max(((self._uct_score(node, child), action, child) \
                for action, child in node.children.items()), key=lambda x: x[0])
        return action, child

    def _uct_score(self, parent, child):
        """
        Computes the uct score of a given node
        """
        return 0


#
# Functions which might be useful for your implementation
#
def apply_new(game: Game, action: int) -> Game:
    game = game.clone()
    game.apply(action)
    return game


def less_then_or_inf(x, y):
    return x < y or (x == y and (x == -float("inf") or x == float("inf")))
